@@ -1,4 +1,6 @@
-__version__ = "5.3.0"
+__version__ = "5.6.2"
+
+from bleak_retry_connector import Allocations
 
 from .advertisement_tracker import (
     TRACKER_BUFFERING_WOBBLE_SECONDS,
@@ -45,6 +47,7 @@ __all__ = [
     "TRACKER_BUFFERING_WOBBLE_SECONDS",
     "UNAVAILABLE_TRACK_SECONDS",
     "AdvertisementTracker",
+    "Allocations",
     "BaseHaRemoteScanner",
     "BaseHaScanner",
     "BluetoothManager",
