@@ -4,8 +4,8 @@ from __future__ import annotations
 
 import asyncio
 import logging
-from collections.abc import Generator
-from contextlib import contextmanager
+from collections.abc import Callable, Generator
+from contextlib import contextmanager, suppress
 from typing import Any
 from unittest.mock import Mock, patch
 
@@ -14,6 +14,7 @@ import pytest
 from bleak.backends.device import BLEDevice
 from bleak.backends.scanner import AdvertisementData
 from bleak.exc import BleakError
+from bleak_retry_connector import Allocations
 from bluetooth_data_tools import monotonic_time_coarse as MONOTONIC_TIME
 
 from habluetooth import BaseHaRemoteScanner, HaBluetoothConnector
@@ -1465,3 +1466,493 @@ async def test_connection_params_no_adapter_idx(
         assert mock_mgmt_ctl.load_conn_params.call_count == 0
 
     cancel_remote()
+
+
+@pytest.mark.asyncio
+async def test_connection_path_scoring_with_slots_and_logging(
+    caplog: pytest.LogCaptureFixture,
+) -> None:
+    """Test connection path scoring and logging reflects slot availability."""
+    from bleak_retry_connector import Allocations
+
+    manager = _get_manager()
+
+    class FakeBleakClientNoConnect(BaseFakeBleakClient):
+        """Fake bleak client that doesn't connect."""
+
+        async def connect(self, *args, **kwargs):
+            """Don't actually connect."""
+            raise BleakError("Test - connection not needed")
+
+    # Create fake connectors
+    fake_connector_1 = HaBluetoothConnector(
+        client=FakeBleakClientNoConnect, source="scanner1", can_connect=lambda: True
+    )
+    fake_connector_2 = HaBluetoothConnector(
+        client=FakeBleakClientNoConnect, source="scanner2", can_connect=lambda: True
+    )
+    fake_connector_3 = HaBluetoothConnector(
+        client=FakeBleakClientNoConnect, source="scanner3", can_connect=lambda: True
+    )
+
+    # Create scanners with different sources
+    scanner1 = FakeScanner("scanner1", "Scanner 1", fake_connector_1, True)
+    scanner2 = FakeScanner("scanner2", "Scanner 2", fake_connector_2, True)
+    scanner3 = FakeScanner("scanner3", "Scanner 3", fake_connector_3, True)
+
+    # Mock get_allocations for each scanner using patch.object
+    with (
+        patch.object(
+            scanner1,
+            "get_allocations",
+            return_value=Allocations(
+                adapter="scanner1",
+                slots=3,
+                free=1,  # Only 1 slot free - should get penalty
+                allocated=["AA:BB:CC:DD:EE:01", "AA:BB:CC:DD:EE:02"],
+            ),
+        ),
+        patch.object(
+            scanner2,
+            "get_allocations",
+            return_value=Allocations(
+                adapter="scanner2",
+                slots=3,
+                free=2,  # 2 slots free - no penalty
+                allocated=["AA:BB:CC:DD:EE:03"],
+            ),
+        ),
+        patch.object(
+            scanner3,
+            "get_allocations",
+            return_value=Allocations(
+                adapter="scanner3",
+                slots=3,
+                free=3,  # All slots free - no penalty
+                allocated=[],
+            ),
+        ),
+    ):
+        cancel1 = manager.async_register_scanner(scanner1)
+        cancel2 = manager.async_register_scanner(scanner2)
+        cancel3 = manager.async_register_scanner(scanner3)
+
+        # Inject advertisements with different RSSI values
+        device1 = generate_ble_device(
+            "00:00:00:00:00:01", "Test Device", {"source": "scanner1"}
+        )
+        adv_data1 = generate_advertisement_data(local_name="Test Device", rssi=-60)
+        scanner1.inject_advertisement(device1, adv_data1)
+
+        device2 = generate_ble_device(
+            "00:00:00:00:00:01", "Test Device", {"source": "scanner2"}
+        )
+        adv_data2 = generate_advertisement_data(local_name="Test Device", rssi=-65)
+        scanner2.inject_advertisement(device2, adv_data2)
+
+        device3 = generate_ble_device(
+            "00:00:00:00:00:01", "Test Device", {"source": "scanner3"}
+        )
+        adv_data3 = generate_advertisement_data(local_name="Test Device", rssi=-70)
+        scanner3.inject_advertisement(device3, adv_data3)
+
+        await asyncio.sleep(0)
+
+        # Try to connect with logging enabled
+        with caplog.at_level(logging.INFO):
+            client = bleak.BleakClient("00:00:00:00:00:01")
+            with suppress(BleakError):
+                await client.connect()
+
+        # Check that the log contains the connection paths with correct scoring
+        log_text = caplog.text
+        assert "Found 3 connection path(s)" in log_text
+
+        # Extract the log line with connection paths
+        for line in caplog.text.splitlines():
+            if "Found 3 connection path(s)" in line:
+                # rssi_diff = best_rssi - second_best_rssi = -60 - (-65) = 5
+                # Scanner 1 has best RSSI (-60) but only 1 slot free, so with penalty:
+                # score = -60 - (5 * 0.76) = -63.8
+                assert "Scanner 1" in line
+                assert "(slots=1/3 free)" in line
+                assert "(score=-63.8)" in line
+
+                # Scanner 2 has RSSI -65 with 2 slots free, no penalty:
+                # score = -65
+                assert "Scanner 2" in line
+                assert "(slots=2/3 free)" in line
+                # Check for both -65 and -65.0
+                assert ("(score=-65)" in line) or ("(score=-65.0)" in line)
+
+                # Scanner 3 has RSSI -70 with all slots free, no penalty:
+                # score = -70
+                assert "Scanner 3" in line
+                assert "(slots=3/3 free)" in line
+                # Check for both -70 and -70.0
+                assert ("(score=-70)" in line) or ("(score=-70.0)" in line)
+
+                # Verify order: Scanner 1 should be first (best score -63.8),
+                # then Scanner 2 (-65), then Scanner 3 (-70)
+                scanner1_pos = line.find("Scanner 1")
+                scanner2_pos = line.find("Scanner 2")
+                scanner3_pos = line.find("Scanner 3")
+
+                assert scanner1_pos < scanner2_pos < scanner3_pos, (
+                    f"Expected Scanner 1 before Scanner 2 before Scanner 3, "
+                    f"but got positions {scanner1_pos}, {scanner2_pos}, {scanner3_pos}"
+                )
+                break
+        else:
+            pytest.fail("Could not find connection path log line")
+
+        cancel1()
+        cancel2()
+        cancel3()
+
+
+@pytest.mark.asyncio
+async def test_connection_path_scoring_no_slots_available(
+    caplog: pytest.LogCaptureFixture,
+) -> None:
+    """Test that scanners with no free slots are excluded."""
+    manager = _get_manager()
+
+    class FakeBleakClientNoConnect(BaseFakeBleakClient):
+        """Fake bleak client that doesn't connect."""
+
+        async def connect(self, *args, **kwargs):
+            """Don't actually connect."""
+            raise BleakError("Test - connection not needed")
+
+    # Create fake connectors
+    fake_connector_1 = HaBluetoothConnector(
+        client=FakeBleakClientNoConnect, source="scanner1", can_connect=lambda: True
+    )
+    fake_connector_2 = HaBluetoothConnector(
+        client=FakeBleakClientNoConnect, source="scanner2", can_connect=lambda: True
+    )
+
+    # Create scanners
+    scanner1 = FakeScanner("scanner1", "Scanner 1", fake_connector_1, True)
+    scanner2 = FakeScanner("scanner2", "Scanner 2", fake_connector_2, True)
+
+    # Mock get_allocations - scanner1 has no free slots
+    with (
+        patch.object(
+            scanner1,
+            "get_allocations",
+            return_value=Allocations(
+                adapter="scanner1",
+                slots=3,
+                free=0,  # No slots available - should be excluded
+                allocated=[
+                    "AA:BB:CC:DD:EE:01",
+                    "AA:BB:CC:DD:EE:02",
+                    "AA:BB:CC:DD:EE:03",
+                ],
+            ),
+        ),
+        patch.object(
+            scanner2,
+            "get_allocations",
+            return_value=Allocations(
+                adapter="scanner2", slots=3, free=3, allocated=[]  # All slots free
+            ),
+        ),
+    ):
+        cancel1 = manager.async_register_scanner(scanner1)
+        cancel2 = manager.async_register_scanner(scanner2)
+
+        # Inject advertisements
+        device1 = generate_ble_device(
+            "00:00:00:00:00:02", "Test Device", {"source": "scanner1"}
+        )
+        adv_data1 = generate_advertisement_data(
+            local_name="Test Device", rssi=-50
+        )  # Better RSSI
+        scanner1.inject_advertisement(device1, adv_data1)
+
+        device2 = generate_ble_device(
+            "00:00:00:00:00:02", "Test Device", {"source": "scanner2"}
+        )
+        adv_data2 = generate_advertisement_data(
+            local_name="Test Device", rssi=-70
+        )  # Worse RSSI
+        scanner2.inject_advertisement(device2, adv_data2)
+
+        await asyncio.sleep(0)
+
+        # Try to connect with logging enabled
+        with caplog.at_level(logging.INFO):
+            client = bleak.BleakClient("00:00:00:00:00:02")
+            with suppress(BleakError):
+                await client.connect()
+
+        # Check that only scanner2 is in the connection paths
+        log_text = caplog.text
+        assert (
+            "Found 1 connection path(s)" in log_text
+            or "Found 2 connection path(s)" in log_text
+        )
+
+        # If both are shown, scanner1 should have bad score (NO_RSSI_VALUE = -127)
+        for line in caplog.text.splitlines():
+            if "connection path(s)" in line:
+                if "Scanner 1" in line:
+                    # Scanner 1 should show 0 free slots and bad score
+                    assert "(slots=0/3 free)" in line
+                    assert "(score=-127)" in line  # NO_RSSI_VALUE
+
+                # Scanner 2 should be present with normal score
+                assert "Scanner 2" in line
+                assert "(slots=3/3 free)" in line
+                # Check for both -70 and -70.0
+                assert ("(score=-70)" in line) or ("(score=-70.0)" in line)
+                break
+
+        cancel1()
+        cancel2()
+
+
+@pytest.mark.asyncio
+async def test_thundering_herd_connection_slots() -> None:
+    """
+    Test thundering herd scenario with limited connection slots.
+
+    Simulates 7 devices trying to connect simultaneously to 3 proxies:
+    - Proxy 1 & 2: Good signal (-60 RSSI), 3 slots each
+    - Proxy 3: Bad signal (-95 RSSI), 3 slots each
+
+    Expected behavior:
+    - First 6 devices should connect to proxy1 and proxy2 (3 each)
+    - 7th device should connect to proxy3 (bad signal) when others are full
+    """
+    from bleak_retry_connector import Allocations
+
+    manager = _get_manager()
+
+    # Track which backend each device connected to
+    connection_tracker = {}
+
+    class FakeBleakClientThunderingHerd(BaseFakeBleakClient):
+        """Fake bleak client for thundering herd test."""
+
+        def __init__(self, address_or_ble_device, *args, **kwargs):
+            """Initialize with tracking."""
+            super().__init__(address_or_ble_device, *args, **kwargs)
+            self._connected = False
+            # Track the device and source
+            if isinstance(address_or_ble_device, BLEDevice):
+                self._address = address_or_ble_device.address
+                self._source = address_or_ble_device.details.get("source")
+            else:
+                self._address = str(address_or_ble_device)
+                self._source = None
+
+        async def connect(self, *args, **kwargs):
+            """Simulate connection and record which backend was used."""
+            # Small delay to simulate connection time
+            await asyncio.sleep(0.01)
+            self._connected = True
+            # Record which backend this device connected to
+            if self._address and self._source:
+                connection_tracker[self._address] = self._source
+            return True
+
+        @property
+        def is_connected(self) -> bool:
+            """Return connection state."""
+            return self._connected
+
+    # Create fake connectors for 3 proxies
+    fake_connector_1 = HaBluetoothConnector(
+        client=FakeBleakClientThunderingHerd,
+        source="proxy1",
+        can_connect=lambda: True,
+    )
+    fake_connector_2 = HaBluetoothConnector(
+        client=FakeBleakClientThunderingHerd,
+        source="proxy2",
+        can_connect=lambda: True,
+    )
+    fake_connector_3 = HaBluetoothConnector(
+        client=FakeBleakClientThunderingHerd,
+        source="proxy3",
+        can_connect=lambda: True,
+    )
+
+    # Create 3 scanners (proxies) with 3 connection slots each
+    proxy1 = FakeScanner("proxy1", "Proxy 1 (Good)", fake_connector_1, True)
+    proxy2 = FakeScanner("proxy2", "Proxy 2 (Good)", fake_connector_2, True)
+    proxy3 = FakeScanner("proxy3", "Proxy 3 (Bad)", fake_connector_3, True)
+
+    # Track actual slot allocations dynamically
+    proxy_allocations: dict[str, set[str]] = {
+        "proxy1": set(),
+        "proxy2": set(),
+        "proxy3": set(),
+    }
+
+    def get_proxy_allocations(proxy_name: str) -> Allocations:
+        """Get allocations for a specific proxy."""
+        allocated = proxy_allocations[proxy_name]
+        return Allocations(
+            adapter=proxy_name,
+            slots=3,
+            free=3 - len(allocated),
+            allocated=list(allocated),
+        )
+
+    # Mock methods to track allocations
+    def make_add_connecting(proxy_name: str) -> Callable[[str], None]:
+        def _add_connecting(addr: str) -> None:
+            proxy_allocations[proxy_name].add(addr)
+
+        return _add_connecting
+
+    def make_finished_connecting(proxy_name: str) -> Callable[[str, bool], None]:
+        def _finished_connecting(addr: str, success: bool) -> None:
+            if not success:
+                proxy_allocations[proxy_name].discard(addr)
+
+        return _finished_connecting
+
+    # Mock get_allocations and connection tracking
+    with (
+        patch.object(
+            proxy1, "get_allocations", lambda: get_proxy_allocations("proxy1")
+        ),
+        patch.object(
+            proxy2, "get_allocations", lambda: get_proxy_allocations("proxy2")
+        ),
+        patch.object(
+            proxy3, "get_allocations", lambda: get_proxy_allocations("proxy3")
+        ),
+        patch.object(proxy1, "_add_connecting", make_add_connecting("proxy1")),
+        patch.object(proxy2, "_add_connecting", make_add_connecting("proxy2")),
+        patch.object(proxy3, "_add_connecting", make_add_connecting("proxy3")),
+        patch.object(
+            proxy1, "_finished_connecting", make_finished_connecting("proxy1")
+        ),
+        patch.object(
+            proxy2, "_finished_connecting", make_finished_connecting("proxy2")
+        ),
+        patch.object(
+            proxy3, "_finished_connecting", make_finished_connecting("proxy3")
+        ),
+    ):
+        cancel1 = manager.async_register_scanner(proxy1)
+        cancel2 = manager.async_register_scanner(proxy2)
+        cancel3 = manager.async_register_scanner(proxy3)
+
+        # Create 7 devices to connect
+        device_addresses = [f"AA:BB:CC:DD:EE:0{i}" for i in range(1, 8)]
+
+        # Inject advertisements for all devices on all proxies
+        for i, address in enumerate(device_addresses, 1):
+            # Good signal on proxy1
+            device1 = generate_ble_device(address, f"Device {i}", {"source": "proxy1"})
+            adv_data1 = generate_advertisement_data(local_name=f"Device {i}", rssi=-60)
+            proxy1.inject_advertisement(device1, adv_data1)
+
+            # Good signal on proxy2 (exactly same as proxy1)
+            device2 = generate_ble_device(address, f"Device {i}", {"source": "proxy2"})
+            adv_data2 = generate_advertisement_data(local_name=f"Device {i}", rssi=-60)
+            proxy2.inject_advertisement(device2, adv_data2)
+
+            # Bad signal on proxy3
+            device3 = generate_ble_device(address, f"Device {i}", {"source": "proxy3"})
+            adv_data3 = generate_advertisement_data(local_name=f"Device {i}", rssi=-95)
+            proxy3.inject_advertisement(device3, adv_data3)
+
+        await asyncio.sleep(0)
+
+        # Clear the connection tracker before starting
+        connection_tracker.clear()
+
+        async def connect_device(address: str) -> tuple[str, str | None]:
+            """Try to connect to a device."""
+            client = bleak.BleakClient(address)
+            try:
+                await client.connect()
+                # The connection tracker should have recorded which backend was used
+                return address, connection_tracker.get(address, "unknown")
+            except BleakError:
+                # Connection failed (no available backend)
+                return address, None
+
+        # Simulate thundering herd - all devices try to connect at once
+        tasks = [connect_device(addr) for addr in device_addresses]
+        results = await asyncio.gather(*tasks, return_exceptions=True)
+
+        # Process results
+        connection_results = {}
+        for result in results:
+            if isinstance(result, tuple):
+                address, proxy = result
+                connection_results[address] = proxy
+
+        # Count connections per proxy
+        proxy1_connections = [
+            addr for addr, p in connection_results.items() if p == "proxy1"
+        ]
+        proxy2_connections = [
+            addr for addr, p in connection_results.items() if p == "proxy2"
+        ]
+        proxy3_connections = [
+            addr for addr, p in connection_results.items() if p == "proxy3"
+        ]
+        failed_connections = [
+            addr for addr, p in connection_results.items() if p is None
+        ]
+
+        # Verify constraints
+        # 1. No proxy should exceed its slot limit
+        assert (
+            len(proxy1_connections) <= 3
+        ), f"Proxy1 exceeded slot limit: {len(proxy1_connections)} > 3"
+        assert (
+            len(proxy2_connections) <= 3
+        ), f"Proxy2 exceeded slot limit: {len(proxy2_connections)} > 3"
+        assert (
+            len(proxy3_connections) <= 3
+        ), f"Proxy3 exceeded slot limit: {len(proxy3_connections)} > 3"
+
+        # 2. Good signal proxies should be preferred and fill up first
+        good_proxy_total = len(proxy1_connections) + len(proxy2_connections)
+        assert (
+            good_proxy_total == 6
+        ), f"Expected exactly 6 connections on good proxies, got {good_proxy_total}"
+
+        # 3. All 7 devices should connect (6 to good proxies, 1 to bad proxy)
+        total_connected = (
+            len(proxy1_connections) + len(proxy2_connections) + len(proxy3_connections)
+        )
+        assert (
+            total_connected == 7
+        ), f"Expected all 7 devices to connect, but only {total_connected} did"
+
+        # 4. The 7th device should go to proxy3 since good ones are full
+        assert (
+            len(proxy3_connections) == 1
+        ), f"Expected exactly 1 connection on proxy3, got {len(proxy3_connections)}"
+
+        # 5. Verify good distribution across proxy1 and proxy2
+        # Both should have roughly equal load (3 connections each)
+        assert (
+            len(proxy1_connections) == 3
+        ), f"Expected proxy1 to have 3 connections, got {len(proxy1_connections)}"
+        assert (
+            len(proxy2_connections) == 3
+        ), f"Expected proxy2 to have 3 connections, got {len(proxy2_connections)}"
+
+        # 6. No connections should fail
+        assert (
+            len(failed_connections) == 0
+        ), f"Expected no failed connections, but {len(failed_connections)} failed"
+
+        # Clean up
+        cancel1()
+        cancel2()
+        cancel3()
